/*
 * robotfindskitten: A Zen simulation
 *
 * Copyright (C) 1997,2000 Leonard Richardson
 *                         leonardr@segfault.org
 *                         http://www.crummy.com/devel/
 *
 * GP32 version 2004 by Thomas Mathys
 * http://members.vantage.ch/killer
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or EXISTANCE OF KITTEN.  See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdlib.h>
#include <string.h>
#include <defines.h>
#include <gpstdlib.h>
#include <gpgraphic.h>
#include <gpfont.h>

#include "console.h"
#include "messages.h"


enum {
	ITEM_FINDKITTEN,
	ITEM_INSTRUCTIONS,
	ITEM_ABOUT,
	ITEM_EXIT,
	NMENUITEMS,
};


enum {
	MENUY = 15,
	STATUSBAR_YOFS = 1,
	STATUSBAR_HEIGHT = 3,
	SCREEN_WIDTH = CON_WIDTH,
	SCREEN_HEIGHT = CON_HEIGHT - STATUSBAR_HEIGHT - 3,
	SCREEN_XOFS = 0,
	SCREEN_YOFS = STATUSBAR_HEIGHT + 2,
};


typedef enum {
	EMPTY = 0,
	ROBOT = 1,
	KITTEN = 2,
	BOGUS = 3,
}ScreenCell;


typedef struct {
	int x;
	int y;
	int color;
	char character;
}screen_object;


static const char title[] = "robotfindskitten";
static const char version[] = VERSION;
static const char *menuitems[NMENUITEMS] = {
	"findkitten!",
	"Instructions",
	"About",
	"Exit",
};
static const char *robot_image[] = {
	"[-]",
	"(+)=C",
	"| |",
	"OOO",
};
static const char *kitten_image[] = {
	"|\\_/|",
	"|o o|__",
	"==*==__\\",
	"C_C_(___)",
};
static const char *instructions[] = {
	"robotfindskitten instructions",
	"-----------------------------",
	"",
	"",
	"In this game, you are robot (#).",
	"Your job is to find kitten.",
	"",
	"This task is complicated by the",
	"existence of various things which",
	"are not kitten.",
	"",
	"Robot must touch items to determine",
	"if they are kitten or not.",
	"",
	"The game ends when robotfindskitten.",
	"",
	"While playing, use joystick to move",
	"robot, select to abort game.",
	"",
	"",
	"Press any button to continue.",
};


screen_object robot;
screen_object kitten;
int num_bogus;
const char *last_bogus_message;
screen_object bogus[MESSAGES];
int bogus_messages[MESSAGES];
int used_messages[MESSAGES];
int screen[SCREEN_WIDTH][SCREEN_HEIGHT];


// Macros for generating numbers in different ranges
#define randx()		(rand() % SCREEN_WIDTH)
#define randy()		(rand() % SCREEN_HEIGHT)
#define randchar()	(rand() % 256)
#define randcolor()	((rand() % (CON_NCOLORS-1)) + 1)


void sleep(u32 ms) {

	u32 start = GpTickCountGet();

	while (abs(GpTickCountGet() - start) < ms);
}


u32 getKey(void) {

	u32 k;

	while (1) {
		sleep(16);
		k = GpKeyGet();
		sleep(16);
		if ( (k == GpKeyGet()) && (k != GPC_VK_NONE) ) {
			break;
		}
	}

	return k;
}


u32 waitKey(void) {
	while (GpKeyGet() != GPC_VK_NONE);
	return getKey();
}


static void printxy(u32 x, u32 y, u32 fgcolor, u32 bgcolor, const char *s) {
	conSetXY(x, y);
	conSetFgColor(fgcolor);
	conSetBgColor(bgcolor);
	conPrint(s);
}


// print centered
static void printcy(u32 y, u32 fgcolor, u32 bgcolor, const char *s) {
	printxy((CON_WIDTH-strlen(s))/2, y, fgcolor, bgcolor, s);
}


// finds the length of the longest string
// in an array of strings.
static u32 findLongestLength(const char *strings[], u32 nstrings) {

	u32 i;
	u32 length = 0;

	for (i=0; i<nstrings; i++) {
		int newlength = strlen(strings[i]);
		if (newlength > length) {
			length = newlength;
		}
	}

	return length;
}


// draw image
//
// x,y		:	x/y position of upper left corner of image on screen
// fg/bgcolor	:	foreground/background color
// height	:	# of lines to draw (length of lines array)
// lines	:	pointers to the lines to draw (each line zero terminated)
static void printImage(u32 x, u32 y, u32 fgcolor, u32 bgcolor, u32 height, const char *lines[]) {

	u32 i;

	for (i=0; i<height; i++) {
		printxy(x, y+i, fgcolor, bgcolor, lines[i]);
	}
}


// display instructions and wait until a key is pressed
static void showInstructions() {

	int i, mx;
	int ninstructions;

	conClr();

	ninstructions = sizeof(instructions) / sizeof(instructions[0]);
	mx = (CON_WIDTH - findLongestLength(instructions, ninstructions)) / 2;

	for (i=0; i<ninstructions; i++) {
		printxy(mx, (CON_HEIGHT-ninstructions)/2+i, CON_WHITE, CON_BLACK, instructions[i]);
	}

	waitKey();
}


// display "about" stuff
static void showUselessInformation() {

	enum {
		FOO = 3
	};

	conClr();

	printcy(FOO+0, CON_WHITE, CON_BLACK, "robotfindskitten");

	printcy(FOO+2, CON_LIGHTGREY, CON_BLACK, "By the illustrious Leonard Richardson");
	printcy(FOO+3, CON_LIGHTGREY, CON_BLACK, "Copyright (C) 1997, 2000");
	printcy(FOO+5, CON_LIGHTGREY, CON_BLACK, "Written originally for the");
	printcy(FOO+6, CON_LIGHTGREY, CON_BLACK, "Nerth Pork robotfindskitten contest.");

	printcy(FOO+10, CON_WHITE, CON_BLACK, "robotfindskitten/GP32");
	printcy(FOO+12, CON_LIGHTGREY, CON_BLACK, "2004 by Thomas Mathys");
	printcy(FOO+13, CON_LIGHTGREY, CON_BLACK, "http://members.vantage.ch/killer");
	printcy(FOO+15, CON_LIGHTGREY, CON_BLACK, "Robot and kitten graphics are borrowed");
	printcy(FOO+16, CON_LIGHTGREY, CON_BLACK, "from Benjamin Molitor's GBA version.");

	printcy(FOO+20, CON_WHITE, CON_BLACK, "http://robotfindskitten.org");

	printcy(FOO+23, CON_LIGHTGREY, CON_BLACK, "Press any button to continue.");

	waitKey();
}


// draw menu
// redraw	:	nonzero = redraw whole menu
//			zero = redraw draw cursor only
// selItem	:	index of currently selected menu item
static void drawMenu(u32 redraw, u32 selItem) {

	static int mx = -1;
	int i;

	if (-1 == mx) {
		mx = (CON_WIDTH - findLongestLength(menuitems, NMENUITEMS)) / 2;
	}

	if (redraw) {
		conClr();

		printImage(5, CON_HEIGHT/8, CON_LIGHTGREY, CON_BLACK, 4, robot_image);
		printImage(26, CON_HEIGHT/8, CON_LIGHTGREY, CON_BLACK, 4, kitten_image);

		printcy(CON_HEIGHT/3, CON_WHITE, CON_BLACK, title);

		// draw menuitems
		for (i=0; i<NMENUITEMS; i++) {
			printxy(mx, MENUY+i*2, CON_LIGHTBLUE, CON_BLACK, menuitems[i]);
		}

		printcy(CON_HEIGHT-2, CON_LIGHTGREY, CON_BLACK, version);
	}

	// redraw cursor
	for (i=0; i<NMENUITEMS; i++) {
		printxy(mx-2, MENUY+i*2, CON_RED, CON_BLACK, " ");
	}
	printxy(mx-2, MENUY+selItem*2, CON_RED, CON_BLACK, "\020");
}


static int validchar(char a) {

	switch (a) {
		case '#':
		case ' ':
		case 0:
		case 255:
			return 0;
			break;
		default:
			return 1;
			break;
	}
}


static void draw(const screen_object *o) {

	conDrawCharAttr(
		SCREEN_XOFS + o->x,
		SCREEN_YOFS + o->y,
		o->character,
		o->color,
		CON_BLACK
	);
}


static void initMessages(void) {

	int i;
	int msg;
	int lastspace;
	int lastwrap;
	int textlen;

	for (msg=0; msg<MESSAGES; msg++) {

		textlen = strlen(messages[msg]);
		lastspace = 0;
		lastwrap = 0;

		for (i=0; i<textlen; i++) {

			if (messages[msg][i] == ' ') {
				lastspace = i;
			}
			if (i-lastwrap >= CON_WIDTH-1) {
				messages[msg][lastspace] = '\n';
				lastwrap = lastspace;
			}
		}
	}
}


static void initArrays(void) {

	int i,x,y;
	screen_object empty;

	empty.x = -1;
	empty.y = -1;
	empty.color = 0;
	empty.character = ' ';

	for (x=0; x<SCREEN_WIDTH; x++) {
		for (y=0; y<SCREEN_HEIGHT; y++) {
			screen[x][y] = EMPTY;
		}
	}

	for (i=0; i<MESSAGES; i++) {
		used_messages[i] = 0;
		bogus_messages[i] = 0;
		bogus[i] = empty;
	}
}


static void initRobot(void) {

	robot.x = randx();
	robot.y = randy();

	robot.character = '#';
	robot.color = CON_LIGHTGREY;
	screen[robot.x][robot.y] = ROBOT;
}


static void initKitten(void) {


	// Assign the kitten a unique position.
	do {
		kitten.x = randx();
		kitten.y = randy();
	} while (screen[kitten.x][kitten.y] != EMPTY);

	// Assign the kitten a character and a color.
	do {
		kitten.character = randchar();
	} while (!(validchar(kitten.character)));

	screen[kitten.x][kitten.y] = KITTEN;

	kitten.color = randcolor();
}


static void initBogus(void) {

	int counter, index;

	for (counter = 0; counter < num_bogus; counter++) {

		// Give it a color.
		bogus[counter].color = randcolor();

		// Give it a character.
		do {
			bogus[counter].character = randchar();
		} while (!(validchar(bogus[counter].character)));

		// Give it a position.
		do {
			bogus[counter].x = randx();
			bogus[counter].y = randy();
		} while (screen[bogus[counter].x][bogus[counter].y] != EMPTY);

		screen[bogus[counter].x][bogus[counter].y] = counter + BOGUS;

		// Find a message for this object.
		do {
			index = rand() % MESSAGES;
		} while (used_messages[index] != 0);

		bogus_messages[counter] = index;
		used_messages[index] = 1;
	}
}


static void initScreen(void) {

	int i;

	conClr();

	conSetXY(0, STATUSBAR_YOFS);
	conSetFgColor(CON_WHITE);
	conSetBgColor(CON_BLACK);
	conPrintf("findkitten!");

	conSetXY(0, STATUSBAR_YOFS+STATUSBAR_HEIGHT);
	for (i=0; i<CON_WIDTH; i++) {
		conPutch('-');
	}

	for (i=0; i<num_bogus; i++) {
		draw(&bogus[i]);
	}

	draw(&robot);
	draw(&kitten);
}


static void printBogusMessage(const char *msg) {

	int x,y;

	if (last_bogus_message != msg) {

		last_bogus_message = msg;

		// clear status bar
		for (y=0; y<STATUSBAR_HEIGHT; y++) {
			for (x=0; x<CON_WIDTH; x++) {
				conDrawCharAttr(x, STATUSBAR_YOFS+y, ' ', CON_BLACK, CON_BLACK);
			}
		}

		// print message
		printxy(0, STATUSBAR_YOFS, CON_WHITE, CON_BLACK, msg);
	}
}


static void victoryAnimation(void) {

	int x;
	const int roboty = CON_HEIGHT/2;
	int heartsvisible;
	u32 start;

	conClr();

	// run towards eachother
	for (x=9; x>=0; x--) {

		// erase old robot/kitten
		conDrawCharAttr(CON_WIDTH/2-1-x-1, roboty, ' ', CON_BLACK, CON_BLACK);
		conDrawCharAttr(CON_WIDTH/2+x+1, roboty, ' ', CON_BLACK, CON_BLACK);

		// draw new robot/kitten
		conDrawCharAttr(CON_WIDTH/2-1-x, roboty, robot.character, robot.color, CON_BLACK);
		conDrawCharAttr(CON_WIDTH/2+x, roboty, kitten.character, kitten.color, CON_BLACK);

		sleep(100);
	}

	printcy(roboty-4, CON_WHITE, CON_BLACK, "You found kitten!");
	printcy(roboty+2, CON_WHITE, CON_BLACK, "Way to go, robot!");
	printcy(roboty+8, CON_LIGHTGREY, CON_BLACK, "Press any button to continue.");

	// flash hearts while waiting for a button
	heartsvisible = 1;
	while (1) {

		if (heartsvisible) {
			printcy(roboty-2, CON_RED, CON_BLACK, "\003\003");
		} else {
			printcy(roboty-2, CON_RED, CON_BLACK, "  ");
		}
		heartsvisible = !heartsvisible;

		start = GpTickCountGet();
		while (abs(GpTickCountGet()-start) < 333) {
			if (GPC_VK_NONE != GpKeyGet()) {
				return;
			}
		}
	}
}


// handle user input
// key		:	pressed key(s)
//
// return value	:	zero = game continues
//			nonzero = game over (won or aborted)
static u32 processInput(u32 key) {

	int check_x = robot.x;
	int check_y = robot.y;

	switch (key) {
		case GPC_VK_LEFT:
			check_x--;
			break;
		case GPC_VK_RIGHT:
			check_x++;
			break;
		case GPC_VK_UP:
			check_y--;
			break;
		case GPC_VK_DOWN:
			check_y++;
			break;
		case GPC_VK_LEFT|GPC_VK_UP:
			check_x--;
			check_y--;
			break;
		case GPC_VK_LEFT|GPC_VK_DOWN:
			check_x--;
			check_y++;
			break;
		case GPC_VK_RIGHT|GPC_VK_UP:
			check_x++;
			check_y--;
			break;
		case GPC_VK_RIGHT|GPC_VK_DOWN:
			check_x++;
			check_y++;
			break;
		case GPC_VK_SELECT:
			return 1;
			break;
	}

	// Check for going off the edge of the screen.
	if ( (check_y < 0) || (check_y >= SCREEN_HEIGHT) ||
	     (check_x < 0) || (check_x >= SCREEN_WIDTH) ) {
		return 0;
	}

	// Check for collision
	if (screen[check_x][check_y] != EMPTY) {
		switch (screen[check_x][check_y]) {
			case ROBOT:
				// We didn't move, or we're stuck in a
				// time warp or something.
				break;
			case KITTEN:
				// Found it!
				victoryAnimation();
				return 1;
				break;
			default:
				// We hit a bogus object; print its message.
				printBogusMessage(messages[bogus_messages[screen[check_x][check_y]-BOGUS]]);
				break;
		}
		return 0;
	}


	// Otherwise, move the robot.
	robot.x = check_x;
	robot.y = check_y;
	return 0;
}


static void game(void) {

	int old_x;
	int old_y;
	u32 key;

	srand(GpTickCountGet());

	num_bogus = 20;
	last_bogus_message = NULL;

	initArrays();
	initRobot();
	initKitten();
	initBogus();
	initScreen();

	while (GpKeyGet() != GPC_VK_NONE) {}

	old_x = robot.x;
	old_y = robot.y;
	while (1) {

		key = getKey();

		if (processInput(key)) {
			return;
		}

      		// Redraw robot, where applicable. We're your station, robot.
		if (!(old_x == robot.x && old_y == robot.y)) {

			// Get rid of the old robot
			conDrawCharAttr(SCREEN_XOFS+old_x, SCREEN_YOFS+old_y,
					' ', CON_BLACK, CON_BLACK);
			screen[old_x][old_y] = EMPTY;

			// Meet the new robot, same as the old robot.
			draw(&robot);
			screen[robot.x][robot.y] = ROBOT;

			old_x = robot.x;
			old_y = robot.y;
		}

		sleep(62);
	}
}


void GpMain(void *arg) {

	int item;

	conInit();
	initMessages();

	item = 0;
	drawMenu(1, item);
	while(1) {

		switch (waitKey()) {

			// next menu item
			case GPC_VK_DOWN:
				item++;
				if (item >= NMENUITEMS) {
					item = 0;
				}
				drawMenu(0, item);
				break;

			// previous menu item
			case GPC_VK_UP:
				item--;
				if (item < 0) {
					item = NMENUITEMS-1;
				}
				drawMenu(0, item);
				break;

			// select
			case GPC_VK_START:
			case GPC_VK_SELECT:
			case GPC_VK_FA:
			case GPC_VK_FB:
				switch (item) {
					case ITEM_FINDKITTEN:
						game();
						drawMenu(1, item);
						break;
					case ITEM_INSTRUCTIONS:
						showInstructions();
						drawMenu(1, item);
						break;
					case ITEM_ABOUT:
						showUselessInformation();
						drawMenu(1, item);
						break;
					case ITEM_EXIT:
						GpAppExit();
						break;
				}
				break;
		}
	}
}
